"""For example you can work with requirements or with versions here. Usually used in 'setup.py'."""

from mypythontools_cicd.packages.packages_internal import (
    get_package_setup_args,
    get_readme,
    get_requirements_files,
    get_requirements,
    get_version,
    personal_setup_args_preset,
    set_version,
    validate_version,
)

__all__ = [
    "get_package_setup_args",
    "get_readme",
    "get_requirements_files",
    "get_requirements",
    "get_version",
    "personal_setup_args_preset",
    "set_version",
    "validate_version",
]



"""Module with functions for 'misc' subpackage."""

from __future__ import annotations
from typing import Sequence


from mypythontools.paths import validate_path, PathLike
from mypythontools.types import validate_sequence

from mypythontools.misc import print_progress
from mypythontools.system import (
    check_script_is_available,
    terminal_do_command,
)

from mypythontools_cicd.project_paths import PROJECT_PATHS


def reformat_with_black(
    root_path: None | PathLike = None, extra_args: Sequence[str] = ("--quiet",), verbose: bool = False
) -> None:
    """Reformat code with black.

    Args:
        root_path (None | PathLike, optional): Root path of project. If None, will be inferred.
            Defaults to None.
        extra_args (Sequence[str], optional): Some extra args for black. Defaults to ("--quiet,").
        verbose (bool, optional): If True, result of terminal command will be printed to console.
            Defaults to False.

    Example:
        >>> reformat_with_black()
    """
    print_progress("Reformatting", verbose)
    check_script_is_available("black", "black")
    validate_sequence(extra_args, "extra_args")

    root_path = validate_path(root_path, "Reformating failed") if root_path else PROJECT_PATHS.root

    terminal_do_command(
        f"black . {' '.join(extra_args)}", cwd=root_path, verbose=verbose, error_header="Formatting failed"
    )



"""Module with functions for 'venvs' subpackage."""

from __future__ import annotations
from typing import Sequence
import platform
import shutil
from pathlib import Path
import sys

from typing_extensions import Literal

from mypythontools.paths import PathLike, is_path_free, WslPath
from mypythontools.misc import print_progress
from mypythontools.system import (
    check_script_is_available,
    get_console_str_with_quotes,
    terminal_do_command,
    SHELL_AND,
    is_wsl,
)

from ..project_paths import PROJECT_PATHS
from ..packages import get_requirements_files
from mypythontools_cicd.project_paths import PROJECT_PATHS


class VenvNotFound(RuntimeError):
    """If some method called on non installed venv."""

    pass


class Venv:
    """You can create new venv or sync it's dependencies.

    Attributes:
        venv_path(Path): Path of venv. E.g. `venv`
        with_wsl(bool, optional): If working with linux venv from linux.

    Example:
        >>> from pathlib import Path
        >>> from mypythontools.system import is_wsl
        ...
        >>> path = "venv/3.10" if platform.system() == "Windows" else "venv/wsl-3.10"
        >>> venv = Venv(path)
        >>> venv.create()  # If already exists, it's skipped
        >>> venv.install_library("colorama==0.3.9")
        >>> "colorama==0.3.9" in venv.list_packages()
        True
        >>> venv.sync_requirements(verbosity=0)  # There ia a 0.4.4 in requirements.txt
        >>> "colorama==0.4.4" in venv.list_packages()
        True
        >>> venv.remove()
        >>> Path(path).exists()
        False
    """

    def __init__(self, venv_path: PathLike, with_wsl: bool = False) -> None:
        """Init the venv class. To create or update it, you can call extra functions.

        Args:
            venv_path(PathLike): Path of venv. E.g. `venv`
            with_wsl(bool, optional): If working with linux venv from linux.

        Raises:
            FileNotFoundError: No folder found on defined path.
        """
        venv_path = Path(venv_path).resolve()
        self.real_path = venv_path

        self.venv_path_console_str = get_console_str_with_quotes(venv_path)
        self.with_wsl = with_wsl

        wsl = is_wsl() or with_wsl

        self.executable_str: str
        """Path to the executables. Can be directly used in terminal. Some libraries cannot use
        ``python -m package`` syntax and therefore it can be called from scripts folder."""

        if platform.system() == "Windows" and not wsl:
            activate_path = venv_path / "Scripts" / "activate.bat"
            self.installed = activate_path.exists()
            self.executable = venv_path / "Scripts" / "python.exe"
            self.create_command = f"python -m venv {self.venv_path_console_str}"
            self.scripts_path = venv_path / "Scripts"
            self.activate_command = get_console_str_with_quotes(activate_path.as_posix())
            self.executable_str = get_console_str_with_quotes((self.executable).as_posix())

        elif platform.system() == "Windows" and with_wsl:
            venv_path = WslPath(venv_path)
            activate_path = venv_path / "bin" / "activate"
            self.installed = (self.real_path / "bin" / "activate").exists()
            self.executable = venv_path / "bin" / "python"
            self.create_command = f"python3 -m venv {self.venv_path_console_str}"
            self.scripts_path = venv_path / "bin"
            self.activate_command = f". {get_console_str_with_quotes(activate_path.wsl_path)}"
            self.executable_str = get_console_str_with_quotes((venv_path / "bin" / "python").wsl_path)

        else:
            activate_path = venv_path / "bin" / "activate"
            self.installed = (venv_path / "bin" / "activate").exists()
            self.executable = venv_path / "bin" / "python"
            self.create_command = f"python3 -m venv {self.venv_path_console_str}"
            self.scripts_path = venv_path / "bin"
            self.activate_command = f". {get_console_str_with_quotes(activate_path)}"
            self.executable_str = get_console_str_with_quotes((venv_path / "bin" / "python"))

        self.venv_path = venv_path
        """Path to venv prefix, e.g. .../venv"""

        self.subprocess_prefix = f"{self.activate_command} {SHELL_AND} {self.executable_str} -m "
        """Run as module, so library can be directly call afterwards. Can be directly used in terminal. It can
        look like this for example::
        
            .../Scripts/activate.bat && .../venv/Scripts/python.exe -m
        """

    def create(self, verbose: bool = False) -> None:
        """Create virtual environment. If it already exists, it will be skipped and nothing happens.

        Args:
            verbose (bool, optional): If True, result of terminal command will be printed to console.
                Defaults to False.
        """
        if not self.installed:
            terminal_do_command(
                self.create_command,
                cwd=PROJECT_PATHS.root.as_posix(),
                shell=True,
                verbose=verbose,
                error_header="Venv creation failed",
                with_wsl=self.with_wsl,
            )
        self.installed = True

    def sync_requirements(
        self,
        requirements: Literal["infer"] | PathLike | Sequence[PathLike] = "infer",
        verbosity: Literal[0, 1, 2] = 1,
    ) -> None:
        """Sync libraries based on requirements. Install missing, remove unnecessary.

        Args:
            requirements (Literal["infer"] | PathLike | Sequence[PathLike], optional): Define what libraries
                will be installed. If "infer", autodetected. Can also be a list of more files e.g
                `["requirements.txt", "requirements_dev.txt"]`. Defaults to "infer".
            verbosity (Literal[0, 1, 2], optional): If 0, prints nothing, if 1, then one line description of what
                happened is printed. If 3, all the results from terminal are printed. Defaults to 1.
        """
        print_progress("Syncing requirements", verbosity > 0)

        self._raise_if_not_installed()

        self.install_library("pip-tools")
        requirements = get_requirements_files(requirements)

        requirements_content = ""

        for i in requirements:
            with open(i, "r") as req:
                requirements_content = requirements_content + "\n" + req.read()

        requirements_all_path = self.venv_path / "requirements_all.in"
        if isinstance(self.venv_path, WslPath):
            requirements_all_console_path_str = get_console_str_with_quotes(requirements_all_path.wsl_path)  # type: ignore
            freezed_requirements_console_path_str = get_console_str_with_quotes(
                (self.venv_path / "requirements.txt").wsl_path
            )
        else:
            freezed_requirements_console_path_str = get_console_str_with_quotes(
                (self.venv_path / "requirements.txt")
            )
            requirements_all_console_path_str = get_console_str_with_quotes(requirements_all_path)

        with open(requirements_all_path, "w") as requirement_libraries:
            requirement_libraries.write(requirements_content)

        pip_compile_command = (
            f"pip-compile {requirements_all_console_path_str} --output-file "
            f"{freezed_requirements_console_path_str} --quiet"
        )

        pip_sync_command = f"pip-sync {freezed_requirements_console_path_str} --quiet"
        sync_commands = {
            pip_compile_command: "Creating joined requirements.txt file failed.",
            pip_sync_command: "Requirements syncing failed.",
        }

        for i, j in sync_commands.items():
            terminal_do_command(
                f"{self.activate_command} {SHELL_AND} {i}",
                verbose=verbosity == 2,
                error_header=j,
                with_wsl=self.with_wsl,
            )

    def list_packages(self) -> str:
        """Get list of installed libraries in the venv.

        The reason why it's meta coded via string parsing and not parsed directly is that it needs to be
        available from other venv as well.
        """
        self._raise_if_not_installed()

        result = terminal_do_command(
            f"{self.activate_command} {SHELL_AND} {self.executable_str} -m pip freeze",
            with_wsl=self.with_wsl,
            verbose=False,
        )

        return result

    def install_library(self, name: str, verbose: bool = False, update: bool = False) -> None:
        """Install package to venv with pip install.

        You can use extras with square brackets.

        Args:
            name (str): Name of installed library.
            verbose (bool, optional): If True, result of terminal command will be printed to console.
                Defaults to False.
            update (bool, optional): Update flag. If True, then latest is installed. If False, and already
                exists, it's skipped. Defaults to False
        """
        self._raise_if_not_installed()

        command = f"{self.activate_command} {SHELL_AND} {self.executable_str} -m pip install {name} {'--update' if update else ''}"
        terminal_do_command(
            command,
            shell=True,
            verbose=verbose,
            error_header="Library installation failed.",
            with_wsl=self.with_wsl,
        )
        terminal_do_command(command, shell=True, verbose=verbose, error_header="Library installation failed.")

    def uninstall_library(self, name: str, verbose: bool = False) -> None:
        """Uninstall package to venv with pip install.

        Args:
            name (str): Name of library to uninstall.
            verbose (bool, optional): If True, result of terminal command will be printed to console.
                Defaults to False.
        """
        self._raise_if_not_installed()

        command = f"{self.activate_command} {SHELL_AND} {self.executable_str} -m pip uninstall {name}"

        terminal_do_command(
            command,
            shell=True,
            verbose=verbose,
            error_header="Library removal failed",
            with_wsl=self.with_wsl,
        )

    def remove(self) -> None:
        """Remove the folder with venv."""
        shutil.rmtree(self.venv_path.as_posix())

    def get_script_path(self, name: str) -> str:
        """Get script path such as pip for example."""
        if platform.system() == "Windows" and not is_wsl():
            return get_console_str_with_quotes(self.scripts_path / f"{name}.exe")
        else:
            return get_console_str_with_quotes(self.scripts_path / name)

    def _raise_if_not_installed(self):
        if not self.installed:

            raise VenvNotFound("Installed venv not found, first create it with 'create' or create multiple.")


def is_venv() -> bool:
    """True if run in venv, False if run in main interpreter directly."""
    return sys.base_prefix.startswith(sys.prefix)


def prepare_venvs(
    path: None | PathLike = "venv",
    versions: Sequence[str] = ["3.7", "3.10", "wsl-3.7", "wsl-3.10"],
    verbosity: Literal[0, 1, 2] = 1,
):
    """This will install virtual environments with defined versions.

    Installation will be skipped if there is already existing folder with content. It is possible tu use wsl
    on windows. You have to install python launcher when using linux or wsl. More about python-launcher
    https://github.com/brettcannon/python-launcher

    Args:
        path (None | PathLike): Where venvs will be stored. If None, cwd() will be used. Defaults to "venv".
        versions (Sequence[str], optional): List of used versions. If you want to use wsl, add `wsl-` prefix
            like for example `wsl-3.7`. Defaults to ["3.7", "3.10", "wsl-3.7", "wsl-3.10"].
        verbosity (Literal[0, 1, 2], optional): If 0, prints nothing, if 1, then one line description of what
            happened is printed. If 3, all the results from terminal are printed. Defaults to 1.
    """
    print_progress("Preparing venvs", verbosity > 0)

    if path is None:
        path = Path.cwd()

    if not isinstance(versions, list):
        raise TypeError("'versions' param has to be list.")

    for version in versions:
        venv_path = Path(f"{path}/{version}")

        if version.startswith("wsl-"):
            wsl = True
            version = version[4:]  # remove "wsl-"
        else:
            wsl = False

        if not is_path_free(venv_path):
            if Venv(venv_path, with_wsl=wsl).installed:
                continue
            else:
                raise RuntimeError(
                    f"There is not empty folder on defined path '{venv_path.as_posix()}' and existing "
                    "virtualenv for current OS not detected there. Clean it first or check the settings "
                    "whether it should be an wsl venv."
                )

        if wsl:
            check_script_is_available(
                "wsl py",
                message=(
                    "Verify whether python launcher is installed. If not, install it from "
                    "https://github.com/brettcannon/python-launcher . \n If it's installed in "
                    "'/home/linuxbrew/.linuxbrew/bin/py' it will be not visible from wsl. "
                    "You can use /usr/local/..."
                ),
            )
        create_command = f"py -{version} -m venv {venv_path.as_posix()}"

        terminal_do_command(
            create_command,
            verbose=verbosity == 2,
            error_header=(
                f"Creating virtual environment for{' wsl ' if wsl else ' '}version {version} failed. "
                "If fails with wsl, try to install 'python3.x-venv' on wsl."
            ),
            with_wsl=wsl,
        )



"""Module with functions for 'tests' subpackage."""

from __future__ import annotations
from typing import Sequence, cast
from pathlib import Path
import sys
import warnings
import os

from typing_extensions import Literal

import mylogging
from mypythontools.paths import validate_path, PathLike, WslPath
from mypythontools.misc import delete_files, print_progress
from mypythontools.config import Config, MyProperty
from mypythontools.system import (
    get_console_str_with_quotes,
    terminal_do_command,
    check_library_is_available,
)

from ..venvs import Venv, prepare_venvs
from ..project_paths import PROJECT_PATHS


class TestConfig(Config):
    """Allow to setup tests."""

    @MyProperty
    @staticmethod
    def tested_path() -> None | PathLike:
        """Define path of tested folder.

        Type:
            None | PathLike

        Default:
            None

        If None, root is used. Root is necessary if using doctest, 'tests' folder not works for doctest.
        """
        return None

    @MyProperty
    @staticmethod
    def run_tests() -> bool:
        """Define whether run tests or not.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def tests_path() -> None | PathLike:
        """If None, tests is used. It means where venv will be stored etc.

        Type:
            None | PathLike

        Default:
            None
        """
        return None

    @MyProperty
    @staticmethod
    def prepare_test_venvs() -> None | list[str]:
        """Create venvs with defined versions.

        Type:
            None | list[str]

        Default:
            ["3.7", "3.10", "wsl-3.7", "wsl-3.10"]
        """
        return ["3.7", "3.10", "wsl-3.7", "wsl-3.10"]

    @MyProperty
    @staticmethod
    def prepare_test_venvs_path() -> PathLike:
        """Prepare venvs in defined path.

        Type:
            str

        Default:
            "tests/venv"
        """
        return "tests/venv"

    @MyProperty
    @staticmethod
    def test_coverage() -> bool:
        """Whether run test coverage plugin. If True, pytest-cov must be installed.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def stop_on_first_error() -> bool:
        """Whether stop on first error.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def virtualenvs() -> Sequence[PathLike]:
        """Virtualenvs used to testing. It's used to be able to test more python versions at once.

        Example:
            ``["tests/venvs/3.7", "tests/venvs/3.10"]``. If you want to use current venv, use `[sys.prefix]`.

        Type:
            None | Sequence[PathLike]

        Default:
            ["tests/venv/3.7", "tests/venv/3.10"]

        If no `virtualenvs` nor `wsl_virtualenvs` is configured, then python that called the function
        will be used.
        """
        return ["tests/venv/3.7", "tests/venv/3.10"]

    @MyProperty
    @staticmethod
    def wsl_virtualenvs() -> Sequence[PathLike]:
        """Define which wsl virtual environments will be tested via wsl.

        Type:
            None | Sequence[PathLike]

        Default:
            ["tests/venv/wsl-3.7", "tests/venv/wsl-3.10"]
        """
        return ["tests/venv/wsl-3.7", "tests/venv/wsl-3.10"]

    @MyProperty
    @staticmethod
    def sync_test_requirements() -> None | Literal["infer"] | PathLike | Sequence[PathLike]:
        """Define whether update libraries versions.

        Type:
            None | Literal["infer"] | PathLike | Sequence[PathLike]

        Default:
            ["requirements.txt"]

        If using `virtualenvs` define what libraries will be installed by path to requirements.txt. Can
        also be a list of more files e.g ``["requirements.txt", "requirements_dev.txt"]``. If "infer",
        auto detected (all requirements).
        """
        return ["requirements.txt"]

    @MyProperty
    @staticmethod
    def verbosity() -> Literal[0, 1, 2]:
        """Define whether print details on errors or keep silent.

        Type:
            Literal[0, 1, 2]

        Default:
            1

        If 0, no details, parameters `-q and `--tb=line` are added. if 1, some details are added --tb=short.
        If 2, more details are printed (default --tb=auto).
        """
        return 1

    @MyProperty
    @staticmethod
    def extra_args() -> None | list:
        """List of args passed to pytest.

        Type:
            None | list

        Default:
            None
        """
        return None


default_test_config = TestConfig()


def run_tests(
    config: TestConfig = default_test_config,
) -> None:
    """Run tests. If any test fails, raise an error.

    This is not supposed for normal testing during development. It's usually part of pipeline an runs just
    before pushing code. It usually runs on more python versions and it's syncing dependencies, so takes much
    more time than testing in IDE.

    Args:
        config (PipelineConfig, optional): TestConfig configuration object. Just import default_test_config
            and use intellisense and help tooltip with description. Defaults to `default_test_config`.

    Raises:
        Exception: If any test fail, it will raise exception (git hook do not continue...).

    Note:
        By default args to quiet mode and no traceback are passed. Usually this just runs automatic tests.
        If some of them fail, it's further analyzed in some other tool in IDE.

    Example:
        ``run_tests(verbosity=2)``
    """
    print_progress("Testing", config.verbosity > 0)

    if not config.run_tests:
        return

    tested_path = (
        validate_path(config.tested_path, "Running tests failed", "tested_path")
        if config.tested_path
        else PROJECT_PATHS.root
    )
    tests_path = (
        validate_path(config.tests_path, "Running tests failed", "tests_path")
        if config.tests_path
        else PROJECT_PATHS.tests
    )

    verbosity = config.verbosity
    verbose = True if verbosity == 2 else False
    inner_verbosity = 2 if verbosity == 2 else 0

    extra_args = config.extra_args if config.extra_args else []

    if config.stop_on_first_error:
        extra_args.append("-x")

    if verbosity == 0:
        extra_args.append("-q")
        extra_args.append("--tb=line")
    elif verbosity == 1:
        extra_args.append("--tb=short")

    all_venvs = [*config.virtualenvs, *[f"wsl-{i}" for i in config.wsl_virtualenvs]]
    if not all_venvs:
        all_venvs = [sys.prefix]

    if config.prepare_test_venvs:
        if verbosity:
            print("\tPreparing test venvs")
        prepare_venvs(path=config.prepare_test_venvs_path, versions=config.prepare_test_venvs, verbosity=0)

    for i, venv in enumerate(all_venvs):
        if venv.startswith("wsl-"):
            wsl = True
            venv = venv[4:]
        else:
            wsl = False

        used_path = tested_path if not wsl else WslPath(tested_path).wsl_path
        tested_path_str = get_console_str_with_quotes(used_path)

        complete_args = (
            "pytest",
            tested_path_str,
            *extra_args,
        )

        test_command = " ".join(complete_args)

        used_command = test_command
        if i == 0:
            if config.test_coverage:
                # Add coverage only to first virtualenv
                xml_path = f"xml:{tests_path / 'coverage.xml'}"
                used_command = used_command + (
                    f" --cov {get_console_str_with_quotes(PROJECT_PATHS.app)} --cov-report "
                    f"{get_console_str_with_quotes(xml_path)}"
                )

        my_venv = Venv(venv, with_wsl=wsl)

        if not my_venv.installed:
            raise RuntimeError(
                f"Defined virtualenv on {my_venv.venv_path} not found. Use 'prepare_test_venvs' or install "
                "venvs manually."
            )

        if config.sync_test_requirements:
            if verbosity:
                print(
                    f"\tSyncing requirements in{' wsl ' if wsl else ' '}venv '{my_venv.venv_path.name}' "
                    "for tests"
                )
            my_venv.sync_requirements(config.sync_test_requirements, verbosity=inner_verbosity)

        # To be able to not install dev requirements in older python venv, pytest is installed.
        # Usually just respond with Requirements already satisfied.
        my_venv.install_library("mypythontools_cicd[tests]")

        used_command = f"{my_venv.activate_command} && {test_command}"

        if verbosity:
            print(f"\tStarting tests with {'wsl' if wsl else ''} venv `{venv}`")

        terminal_do_command(
            used_command,
            cwd=tested_path.as_posix(),
            verbose=verbose,
            error_header="Tests failed.",
            with_wsl=wsl,
        )

    if config.test_coverage:
        delete_files(".coverage")


def setup_tests(
    generate_readme_tests: bool = True,
    matplotlib_test_backend: bool = False,
    set_numpy_random_seed: int | None = 2,
) -> None:
    """Add paths to be able to import local version of library as well as other test files.

    Value Mylogging.config.colorize = 0 changed globally.

    Note:
        Function expect `tests` folder on root. If not, test folder will not be added to sys path and
        imports from tests will not work.

    Args:
        generate_readme_tests (bool, optional): If True, generate new tests from readme if there are
            new changes. Defaults to True.
        matplotlib_test_backend (bool, optional): If using matlplotlib, it need to be
            closed to continue tests. Change backend to agg. Defaults to False.
        set_numpy_random_seed (int | None): If using numpy random numbers, it will be each time the same.
            Numpy is not in requirements, so it need to be installed. Defaults to 2.

    """
    mylogging.config.colorize = False

    PROJECT_PATHS.add_root_to_sys_path()

    # Find paths and add to sys.path to be able to import local modules
    test_dir_path = PROJECT_PATHS.tests

    if test_dir_path.as_posix() not in sys.path:
        sys.path.insert(0, test_dir_path.as_posix())

    if matplotlib_test_backend:
        check_library_is_available("matplotlib")
        import matplotlib

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            matplotlib.use("agg")

    if generate_readme_tests:
        add_readme_tests()

    if set_numpy_random_seed:
        import numpy as np

        np.random.seed(2)


def add_readme_tests(readme_path: None | PathLike = None, tests_folder_path: None | PathLike = None) -> None:
    """Generate pytest tests script file from README.md and save it to tests folder.

    Can be called from conftest.

    Args:
        readme_path (None | PathLike, optional): If None, autodetected (README.md, Readme.md or readme.md
            on root). Defaults to None.
        tests_folder_path (None | PathLike, optional): If None, autodetected (if root / tests).
            Defaults to None.

    Raises:
        FileNotFoundError: If Readme not found.

    Example:
        >>> add_readme_tests()

        Readme tests found.

    Note:
        Only blocks with python defined syntax will be evaluated. Example::

            ```python
            import numpy
            ```

        If you want to import modules and use some global variables, add ``<!--phmdoctest-setup-->`` directive
        before block with setup code.
        If you want to skip some test, add ``<!--phmdoctest-mark.skip-->``
    """
    readme_path = (
        validate_path(readme_path, "'add_readme_tests' failed", "README")
        if readme_path
        else PROJECT_PATHS.readme
    )
    tests_folder_path = (
        validate_path(tests_folder_path, "'add_readme_tests' failed", "tests_folder")
        if tests_folder_path
        else PROJECT_PATHS.tests
    )

    readme_date_modified = str(readme_path.stat().st_mtime).split(".", maxsplit=1)[0]  # Keep only seconds
    readme_tests_name = f"test_readme_generated-{readme_date_modified}.py"

    test_file_path = tests_folder_path / readme_tests_name

    # File not changed from last tests
    if test_file_path.exists():
        return

    for i in tests_folder_path.glob("*"):
        if i.name.startswith("test_readme_generated"):
            i.unlink()

    python_path = get_console_str_with_quotes(sys.executable)
    readme = get_console_str_with_quotes(readme_path)
    output = get_console_str_with_quotes(test_file_path)

    generate_readme_test_command = f"{python_path} -m phmdoctest {readme} --outfile {output}"

    terminal_do_command(generate_readme_test_command, error_header="Readme test creation failed")


def deactivate_test_settings() -> None:
    """Deactivate functionality from setup_tests.

    Sometimes you want to run test just in normal mode (enable plots etc.). Usually at the end of
    test file in ``if __name__ = "__main__":`` block.
    """
    mylogging.config.colorize = True

    if "matplotlib" in sys.modules:

        import matplotlib
        from importlib import reload

        reload(matplotlib)



"""Module with functions around testing. You can run tests including doctest with generating coverage,
you can generate tests from readme or you can configure tests in conftest with single call.

Attributes:

    default_test_config (TestConfig): Default values for tests config. If something changes here, it will
        change in all the repos. You can edit any values in pipeline. Intellisense and help tooltip should
        help.
    """

from mypythontools_cicd.tests.tests_internal import (
    add_readme_tests,
    deactivate_test_settings,
    default_test_config,
    run_tests,
    setup_tests,
    TestConfig,
)

__all__ = [
    "add_readme_tests",
    "deactivate_test_settings",
    "default_test_config",
    "run_tests",
    "setup_tests",
    "TestConfig",
]



"""Module with functions for 'packages' subpackage."""

from __future__ import annotations
from typing import Sequence, Any
from pathlib import Path
import os
import pkg_resources
import re

from typing_extensions import Literal
from mypythontools.paths import validate_path, PathLike

from mypythontools_cicd.project_paths import PROJECT_PATHS

# Lazy import
# from setuptools import find_packages


def get_requirements_files(requirements: Literal["infer"] | PathLike | Sequence[PathLike]) -> list[Path]:
    """Consolidate requirements into list of paths.

    Args:
        requirements (Literal["infer"] | PathLike | Sequence[PathLike]): E.g. ["requirements.txt",
            "requirements_dev.txt"]. If 'infer', then every file where requirements is in the name is used.

    Returns:
        list[Path]: List of paths to requirements files.

    Raises:
        RuntimeError: If no requirements found with 'infer'.
    """

    if requirements == "infer":

        requirements_parsed = []

        for i in PROJECT_PATHS.root.glob("*"):
            if "requirements" in i.as_posix().lower() and i.suffix == ".txt":
                requirements_parsed.append(i)

        if not requirements_parsed:
            raise RuntimeError("No requirements found.")

    else:
        if isinstance(requirements, (Path, str, os.PathLike)):
            requirements = [requirements]

        requirements_parsed = [
            validate_path(req, "'get_requirements_files' failed", "Requirements file") for req in requirements
        ]

    return requirements_parsed


def get_requirements(paths: Literal["infer"] | PathLike | Sequence[PathLike]) -> list[str]:
    """Get requirements into variable usually used in setup.py.

    Args:
        paths (Literal["infer"] | PathLike | Sequence[PathLike]): E.g. ["requirements.txt",
            "requirements_dev.txt"]. If 'infer', then every file where requirements is in the name is used.

    Returns:
        list[str]: List of requirements
    """
    files = get_requirements_files(paths)
    all_requirements = []

    for file in files:
        with open(file) as f:
            requirements = f.readlines()
        requirements = [i.strip("\r\n") for i in requirements if i.strip("\r\n")]

        # narrow_requirements calls recursively get_requirements so it can have another -r reference
        requirements = narrow_requirements(requirements)
        requirements = [str(requirement) for requirement in pkg_resources.parse_requirements(requirements)]
        all_requirements.extend(requirements)

    return all_requirements


def narrow_requirements(requirements: list[str], path: PathLike = "requirements") -> list[str]:
    """If there is another requirements file reference in requirements.txt, this will replace it with its
    requirements it contains.

    Args:
        requirements (list[str]): List of requirements. You can parse it from file with `get_requirements`.
        path (PathLike, optional): Path with referenced files. Defaults to "requirements".

    Returns:
        list[str]: List of plain pep 508 compatible requirements.
    """
    path = validate_path(path, "'narrow_requirements' failed", "File referenced in requirements file")

    for i in requirements:
        if i.startswith("-r "):
            req_name = i[3:].strip("\r\n")
            requirements.extend(get_requirements(path / req_name))
    requirements = [i for i in requirements if not i.startswith("-r ")]
    return requirements


def validate_version(version: str):
    """Check whether parsed version is valid.

    Args:
        version (str): E.g "1.0.1"

    Returns:
        bool: Whether is valid.
    """
    return version == "increment" or (
        len(version.split(".")) == 3 and all([i.isdecimal() for i in version.split(".")])
    )


def get_readme() -> str:
    """Get README content into variable usually used in setup.py.

    Returns:
        str: Content.
    """
    with open(PROJECT_PATHS.readme.as_posix()) as readme_file:
        readme = readme_file.read()
    return readme


def set_version(
    version: str = "increment",
    init_path: None | PathLike = None,
) -> None:
    """Change your version in your `__init__.py` file.

    Args:
        version (str, optional): Form that is used in `__init__`, so for example "1.2.3". Do not use 'v'
            appendix. If version is 'increment', it will increment your `__version__` in you `__init__.py` by
            0.0.1. Defaults to "increment".
        init_path (None | PathLike, optional): Path of file where `__version__` is defined.
            Usually `__init__.py`. If None, will be inferred. Defaults to None.

    Raises:
        ValueError: If no `__version__` is find.
    """
    init_path = (
        validate_path(init_path, "Setting version failed", "__init__.py") if init_path else PROJECT_PATHS.init
    )

    is_valid = validate_version(version)

    if not is_valid:
        raise ValueError(
            "Version not validated. Version has to be of form '1.2.3'. Three digits and two dots. "
            f"You used {version}"
        )

    with open(init_path.as_posix(), "r") as init_file:

        list_of_lines = init_file.readlines()

        found = False

        for i, j in enumerate(list_of_lines):
            if j.startswith("__version__"):

                found = True

                delimiter = '"' if '"' in j else "'"
                delimited = j.split(delimiter)

                if version == "increment":
                    version_list = delimited[1].split(".")
                    version_list[2] = str(int(version_list[2]) + 1)
                    delimited[1] = ".".join(version_list)

                else:
                    delimited[1] = version

                list_of_lines[i] = delimiter.join(delimited)
                break

        if not found:
            raise ValueError("__version__ variable not found in __init__.py. Try set init.")

    with open(init_path.as_posix(), "w") as init_file:

        init_file.writelines(list_of_lines)


def get_version(init_path: None | PathLike = None) -> str:
    """Get version info from `__init__.py` file.

    Args:
        init_path (None | PathLike, optional): Path to `__init__.py` file. If None, will be inferred.
            Defaults to None.

    Returns:
        str: String of version from `__init__.py`.

    Raises:
        ValueError: If no `__version__` is find. Try set init_path...

    Example:
        >>> version = get_version()
        >>> len(version.split(".")) == 3 and all([i.isdecimal() for i in version.split(".")])
        True
    """
    init_path = init_path if init_path else PROJECT_PATHS.init
    try:
        init_path = validate_path(init_path)
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Setting version failed. '__init__.py' not found on '{init_path}'."
        ) from None

    with open(init_path.as_posix(), "r") as init_file:
        version = re.findall('__version__ = "(.*)"', init_file.read())[0]

    if validate_version(version):
        return version
    else:
        raise RuntimeError("Version not found in __init__.py")


def get_package_setup_args(
    name: str, development_status: Literal["alpha", "beta", "stable"] = "alpha"
) -> dict[str, Any]:
    """Get setup args, that usually repeats across projects.

    Args:
        name (str): Name used to generate some params.
        development_status (Literal['alpha', 'beta', 'stable']): Project phase. Defaults to 'alpha'.

    Returns:
        dict[str, Any]: Attributes used in `setup.py`. E.g. license, platform
            or long_description_content_type.
    """
    from setuptools import find_packages

    status = {
        "alpha": "Development Status :: 3 - Alpha",
        "beta": "Development Status :: 4 - Beta",
        "stable": "Development Status :: 5 - Production/Stable",
    }[development_status]

    classifiers = [
        status,
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]

    return {
        "license": "mit",
        "include_package_data": True,
        "long_description_content_type": "text/markdown",
        "packages": find_packages(exclude=("tests**",)),
        "platforms": "any",
        "python_requires": ">=3.7",
        "url": f"https://github.com/Malachov/{name}",
        "name": name,
        "version": get_version(),
        "classifiers": classifiers,
    }


personal_setup_args_preset = {
    "author_email": "malachovd@seznam.cz",
    "author": "Daniel Malachov",
}



"""Helpers with git and github functionality."""

from mypythontools_cicd.docs.docs_internal import docs_regenerate, generate_readme_from_init

__all__ = ["docs_regenerate", "generate_readme_from_init"]



"""Miscellaneous functions that are too small to have own subpackage like for example formatting with black.
"""

from mypythontools_cicd.misc.misc_internal import reformat_with_black

__all__ = ["reformat_with_black"]



"""Module with functions for 'cicd' subpackage."""

from __future__ import annotations
from typing import Sequence
import os
import sys

from typing_extensions import Literal

from mypythontools.config import Config, MyProperty
from mypythontools.misc import GLOBAL_VARS, EMOJIS, print_progress
from mypythontools.paths import PathLike

from .. import git
from .. import venvs
from .. import tests
from ..deploy import deploy_to_pypi
from ..docs import docs_regenerate
from ..misc import reformat_with_black

from ..packages import (
    get_version,
    set_version,
)


class PipelineConfig(Config):
    """Allow to setup CICD pipeline."""

    def __init__(self) -> None:
        """Create subconfigs."""
        self.test: tests.TestConfig = tests.default_test_config

    @MyProperty
    @staticmethod
    def do_only() -> Literal[
        None,
        "prepare_venvs",
        "reformat",
        "test",
        "docs",
        "sync_test_requirements",
        "git_commit_all",
        "git_push",
        "deploy",
    ]:
        """Run just single function from pipeline, ignore the others.

        Type:
            Literal[
                None, "prepare_venvs", "reformat", "test", "docs", "sync_test_requirements", "git_commit_all",
                "git_push", "deploy"
            ]

        Default:
            None

        Reason for why to call it form here and not directly is to be able to use sys args or single command
        line entrypoint.
        """
        return None

    @MyProperty
    @staticmethod
    def prepare_venvs() -> None | list[str]:
        """Create venvs with defined versions.

        Type:
            None | list[str]

        Default:
            None
        """
        return None

    @MyProperty
    @staticmethod
    def prepare_venvs_path() -> PathLike:
        """Prepare venvs in defined path.

        Type:
            str

        Default:
            "venv"
        """
        return "venv"

    @MyProperty
    @staticmethod
    def reformat() -> bool:
        """Reformat all python files with black. Setup parameters in pyproject.toml.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def version() -> None | str:
        """Overwrite __version__ in __init__.py.

        Type:
            str

        Default:
            'increment'.

        Version has to be in format like '1.0.3' three digits and two dots. If 'None', nothing will happen. If
        'increment', than it will be updated by 0.0.1..
        """
        return "increment"

    @MyProperty
    @staticmethod
    def docs() -> bool:
        """Define whether generate sphinx apidoc and generate rst files for documentation.

        Type:
            bool

        Default:
            True

        Some files in docs source can be deleted - check `docs` docstrings for details.
        """
        return True

    @MyProperty
    @staticmethod
    def sync_requirements() -> None | Literal["infer"] | PathLike | Sequence[PathLike]:
        """Check requirements.txt and update all the libraries.

        Type:
            None | Literal["infer"] | PathLike | Sequence[PathLike]

        Default:
            None

        You can use path to requirements, list of paths or bool value. If True, then path is inferred.
        """
        return None

    @MyProperty
    @staticmethod
    def git_commit_all() -> None | str:
        """Whether take all the changes in repository and create a commit with these changes.

        Note:
            !!! Be cautious here if using with `git_push` !!!

        Type:
            None | str

        Default:
            'New commit'
        """
        return "New commit"

    @MyProperty
    @staticmethod
    def git_push() -> bool:
        """Whether push to repository.

        Type:
            bool

        Default:
            True
        """
        return True

    @MyProperty
    @staticmethod
    def tag() -> str:
        """Tag. E.g 'v1.1.2'. If '__version__', get the version.

        Type:
            str

        Default:
            '__version__'
        """
        return "__version__"

    @MyProperty
    @staticmethod
    def tag_message() -> str:
        """Tag message.

        Type:
            str

        Default:
            'New version'
        """
        return "New version"

    @MyProperty
    @staticmethod
    def deploy() -> bool:
        """Deploy to PYPI.

        `TWINE_USERNAME` and `TWINE_PASSWORD` are used for authorization.

        Type:
            bool

        Default:
            False
        """
        return False

    @MyProperty
    @staticmethod
    def allowed_branches() -> None | Sequence[str]:
        """Pipeline runs only on defined branches.

        Type:
            None | Sequence[str]

        Default:
            ["master", "main"]
        """
        return ["master", "main"]

    @MyProperty
    @staticmethod
    def verbosity() -> Literal[0, 1, 2]:
        """Pipeline runs only on defined branches.

        Type:
            Literal[0, 1, 2]

        Default:
            1
        """
        return 1


default_pipeline_config = PipelineConfig()


def cicd_pipeline(
    config: PipelineConfig = default_pipeline_config,
) -> None:
    """Run pipeline for pushing and deploying an app or a package.

    Can run tests, generate rst files for sphinx docs, push to github and deploy to pypi. All params can be
    configured not only with function params, but also from command line with params and therefore callable
    from terminal and optimal to run from IDE (for example with creating simple VS Code task).

    Some function suppose some project structure (where are the docs, where is `__init__.py` etc.).
    If you are issuing some error, try functions directly, find necessary paths in parameters
    and set paths that are necessary in paths module.

    Note:
        Beware, that by default, it creates a commit and add all the changes, not only the staged ones.

    When using sys args for boolean values, always define True or False.

    There is command line entrypoint called `mypythontools_cicd`. After mypythontools is installed, you can
    use it in terminal like::

        mypythontools_cicd --do_only reformat

    Args:
        config (PipelineConfig, optional): PipelineConfig object with CICD pipeline configuration. Just import
            default_pipeline_config and use intellisense and help tooltip with description. It is also
            possible to configure all the params with CLI args from terminal.
            Defaults to `default_pipeline_config`.

    Example:
        Recommended use is from IDE (for example with Tasks in VS Code). Check utils docs for how to use it.
        You can also use it from python... ::

            from mypythontools_cicd.cicd import cicd_pipeline, default_pipeline_config

            default_pipeline_config.deploy = True
            # default_pipeline_config.do_only = ""


            if __name__ == "__main__":
                # All the parameters can be overwritten via CLI args
                cicd_pipeline(config=default_pipeline_config)

        It's also possible to use CLI and configure it via args. This example just push repo to PyPi. ::

            python path-to-project/utils/push_script.py --do_only deploy

    Another way how to run it is to use IDE. For example in VS Code you can use Tasks. Check `cicd`
    docs for examples.
    """
    if not GLOBAL_VARS.is_tested:
        config.with_argparse()

    do_only = config.do_only

    # If do_only change subconfig value and not config, change it
    if do_only == "test":
        do_only = "run_tests"

    if do_only:
        do_only_value = config[do_only]
        config.update(
            {
                "prepare_venvs": None,
                "reformat": False,
                "run_tests": False,
                "docs": False,
                "sync_requirements": None,
                "git_commit_all": None,
                "git_push": False,
                "deploy": False,
                "version": None,
            }
        )
        config.update({do_only: do_only_value})

        if config.verbosity == 1:
            config.verbosity = 0

    verbosity = config.verbosity
    verbose = verbosity == 2
    progress_is_printed = verbosity > 0

    if config.prepare_venvs:
        venvs.prepare_venvs(
            path=config.prepare_venvs_path, versions=config.prepare_venvs, verbosity=verbosity
        )

    if config.allowed_branches:
        git.check_branch(config.allowed_branches)

    # Do some checks before run pipeline so not need to rollback eventually
    if config.deploy:
        usr = os.environ.get("TWINE_USERNAME")
        pas = os.environ.get("TWINE_PASSWORD")

        if not usr or not pas:
            raise KeyError("Setup env vars TWINE_USERNAME and TWINE_PASSWORD to use deploy.")

    if config.sync_requirements:
        if not venvs.is_venv:
            raise RuntimeError("'sync_requirements' available only if using virtualenv.")
        my_venv = venvs.Venv(sys.prefix)
        my_venv.sync_requirements(config.sync_requirements, verbosity=verbosity)

    if config.test:
        tests.run_tests(config.test)

    if config.reformat:
        reformat_with_black()

    if config.version and config.version != "None":
        print_progress("Setting version", progress_is_printed)
        original_version = get_version()
        set_version(config.version)

    try:
        if config.docs:
            docs_regenerate(verbosity=verbosity)

        if config.git_commit_all:
            git.commit_all(config.git_commit_all, verbosity=verbosity)

        if config.git_push:
            git.push(
                tag=config.tag,
                tag_message=config.tag_message,
                verbosity=verbosity,
            )

    except Exception as err:  # pylint: disable=broad-except
        if config.version:
            set_version(original_version)  # type: ignore

        raise RuntimeError(
            f"{3 * EMOJIS.DISAPPOINTMENT} Utils pipeline failed {3 * EMOJIS.DISAPPOINTMENT} \n\n"
            "Original version restored. Nothing was pushed to repo, you can restart pipeline."
        ) from err

    try:
        if config.deploy:
            deploy_to_pypi(verbosity=verbosity)

    except Exception as err:  # pylint: disable=broad-except
        raise RuntimeError(
            f"{3 * EMOJIS.DISAPPOINTMENT} Deploy failed {3 * EMOJIS.DISAPPOINTMENT} \n\n"
            "Already pushed to repository. Deploy manually. Version already changed.",
        ) from err

    print_progress(f"{3 * EMOJIS.PARTY} Finished {3 * EMOJIS.PARTY}", True)



"""Module with functions for 'doc' subpackage."""

from __future__ import annotations
from typing import Sequence
import ast

from typing_extensions import Literal

from mypythontools.paths import validate_path, PathLike
from mypythontools.types import validate_sequence

from mypythontools.misc import delete_files, print_progress
from mypythontools.system import (
    check_script_is_available,
    get_console_str_with_quotes,
    terminal_do_command,
    SHELL_AND,
)

from mypythontools_cicd.project_paths import PROJECT_PATHS


def docs_regenerate(
    docs_path: None | PathLike = None,
    build_locally: bool = True,
    git_add: bool = True,
    keep: Sequence[PathLike] = (),
    ignore: Sequence[PathLike] = ("modules.rst", "**/*internal.py"),
    verbosity: Literal[0, 1, 2] = 1,
) -> None:
    """Generate all rst files necessary for sphinx documentation generation with sphinx-apidoc.

    It automatically delete rst files from removed or renamed files.

    Note:
        All the files except ['conf.py', 'index.rst', '_static', '_templates', 'content/**'], and files in
        'keep' parameter will be deleted!!! Because if some files would be deleted or
        renamed, rst would stay and html was generated. If you have some extra files or folders in docs
        source, add it to content folder or to 'keep' parameter.

    Function suppose sphinx build and source in separate folders...

    Args:
        docs_path (None | PathLike, optional): Where source folder is. If None, will be inferred.
            Defaults to None.
        build_locally (bool, optional): If true, build folder with html files locally.
            Defaults to True.
        git_add (bool, optional): Whether to add generated files to stage. False mostly for
            testing reasons. Defaults to True.
        keep (Sequence[PathLike], optional): List of files and folder names that will not be
            deleted. Deletion is because if some file would be renamed or deleted, rst docs would still stay.
            Glob-style patterns can be used, but it's not recursive, but only first level of source folder is
            used. Defaults to None.
        ignore (Sequence[PathLike], optional): Whether ignore some files from generated rst files. For example
            It can be python modules that will be ignored or it can be rst files created, that will be
            deleted. to have no errors in sphinx build for unused modules, or for internal modules. Glob-style
            patterns can be used. Defaults to ("modules.rst", "**/*_.py")
        verbosity (Literal[0, 1, 2], optional): If 0, prints nothing, if 1, then one line description of what
            happened is printed. If 3, all the results from terminal are printed. Defaults to 1.

    Note:
        Function suppose structure of docs like::

            -- docs
            -- -- source
            -- -- -- conf.py
            -- -- make.bat
    """
    print_progress("Sphinx docs generation", verbosity > 0)
    check_script_is_available("sphinx-apidoc", "sphinx")

    validate_sequence(keep, keep)
    validate_sequence(ignore, ignore)

    verbose = verbosity == 2

    docs_path = (
        validate_path(docs_path, "Docs generation with 'docs_regenerate' failed", "Docs path")
        if docs_path
        else PROJECT_PATHS.docs
    )

    docs_source_path = docs_path / "source"

    source_path = PROJECT_PATHS.app
    source_console_path = get_console_str_with_quotes(source_path)

    keep = [
        *keep,
        "conf.py",
        "index.rst",
        "_static",
        "_templates",
        "content",
    ]
    ignore_list = [*ignore]

    ignored = " "

    for i in ignore_list:
        for file in source_path.rglob(str(i)):
            ignored = ignored + f"{get_console_str_with_quotes(file)} "

    for file in docs_source_path.iterdir():
        if not any((file.match(str(pattern)) for pattern in keep)):
            delete_files(file)

    apidoc_command = (
        f"sphinx-apidoc --module-first --force --separate -o source {source_console_path} {ignored}"
    )

    terminal_do_command(
        apidoc_command, cwd=docs_path, verbose=verbose, error_header="Docs sphinx-apidoc failed."
    )

    if ignore_list:
        for file in docs_source_path.iterdir():
            if any((file.match(str(pattern)) for pattern in ignore_list)):
                delete_files(file)

    if build_locally:
        terminal_do_command(
            f"make clean {SHELL_AND} make html",
            cwd=docs_path,
            verbose=verbose,
            error_header="Sphinx build failed.",
            shell=True,
        )

    if git_add:
        terminal_do_command("git add docs", cwd=PROJECT_PATHS.root.as_posix(), verbose=verbose)


def generate_readme_from_init(git_add: bool = True) -> None:
    """Generate README file from `__init__.py` docstrings.

    Because i had very similar things in main `__init__.py` and in readme. It was to maintain news
    in code. For better simplicity i prefer write docs once and then generate. One code, two use cases.

    Why `__init__`? - Because in IDE on mouseover developers can see help.
    Why README.md? - Good for github.com

    Args:
        git_add (bool, optional): Whether to add generated files to stage. False mostly
            for testing reasons. Defaults to True.
    """
    with open(PROJECT_PATHS.init.as_posix()) as init_file:
        file_contents = init_file.read()
    module = ast.parse(file_contents)
    docstrings = ast.get_docstring(module)

    if docstrings is None:
        docstrings = ""

    with open(PROJECT_PATHS.readme.as_posix(), "w") as file:
        file.write(docstrings)

    if git_add:
        terminal_do_command(f"git add {PROJECT_PATHS.readme}", cwd=PROJECT_PATHS.root.as_posix())



"""Run cicd functions from other modules in particular order and provide configuration.

This module can be used for example in running deploy pipelines or githooks
(some code automatically executed before commit). This module can run the tests,
edit library version, generate rst files for docs, push to git or deploy app to pypi.

All of that can be done with one function call - with `cicd_pipeline` function that
run other functions, or you can use functions separately. 

Attributes:
    default_pipeline_config (PipelineConfig): Default values for pipeline. If something changes here, it will
        change in all the repos. You can edit any values in pipeline. Intellisense and help tooltip should
        help.

Examples:
=========

    **From python script**

    Create folder utils, create `push_script.py` inside, add::

        from mypythontools_cicd.cicd import cicd_pipeline, default_pipeline_config

        if __name__ == "__main__":
            cicd_pipeline(config=default_pipeline_config)

    **With terminal command**

    Run code like this in terminal::

        mypythontools_cicd --test True --deploy True

    If you need to ensure that it will run from particular venv, prepend ``path_to_python/python.exe -m``

    **VS Code Task example**

    You can push changes with single click with all the hooks displaying results in
    your terminal. All params changing every push (like git message or tag) can
    be configured on the beginning and therefore you don't need to wait for test finish.
    Default values can be also used, so in small starting projects, push is actually very fast.

    Here you can find all the tasks with with interactive version and git message config

    https://github.com/Malachov/Software-settings/blob/master/VS-Code/tasks.json

    Add the tasks to global tasks.json. Here is example of one such a task ::

        {
          "version": "2.0.0",
          "tasks": [
            {
                "label": "Push to PyPi",
                "command": "${command:python.interpreterPath}",
                "args": ["-m", "mypythontools_cicd", "--do_only", "reformat"]
            },
          ]
        }

    **Git hooks example**

    Create folder git_hooks with git hook file - for pre commit name must be `pre-commit`
    (with no extension). Hooks in git folder are gitignored by default (and hooks is not visible
    on first sight).

    Then add hook to git settings - run in terminal (last arg is path (created folder))::

        $ git config core.hooksPath git_hooks

    In created folder on first two lines copy this::

        #!/usr/bin/env python
        # -*- coding: UTF-8 -*-

    Then just import any function from here and call with desired params. E.g.
"""
from mypythontools_cicd.cicd.cicd_internal import (
    default_pipeline_config,
    PipelineConfig,
    cicd_pipeline,
)

__all__ = ["default_pipeline_config", "PipelineConfig", "cicd_pipeline"]



"""Module that will help with virtual environments so it's possible to work with venv from python.
You can create, delete or update dependencies."""

from mypythontools_cicd.venvs.venvs_internal import Venv, is_venv, prepare_venvs

__all__ = ["Venv", "is_venv", "prepare_venvs"]



"""Run pipeline if using ``python -m mypythontools_cicd``.

This can be better than console_script when want to ensure that libraries from current venv will be used.
"""

from mypythontools_cicd.cicd import cicd_pipeline

if __name__ == "__main__":
    # Function is configured via sys args
    cicd_pipeline()



"""Install the package."""
from setuptools import setup, Require
from pathlib import Path
import sys

root_path_str = Path(__file__).parent.as_posix()

if root_path_str not in sys.path:
    sys.path.insert(0, root_path_str)

from mypythontools_cicd import packages


if __name__ == "__main__":

    extras_requirements = {
        i: packages.get_requirements(f"requirements/requirements_extras_{i}.txt")
        for i in [
            "all",
            "build",
            "cicd",
            "deploy",
            "dev",
            "docs",
            "git",
            "misc",
            "tests",
            "venvs",
        ]
    }

    setup(
        description="Some tools/functions/snippets used across projects.",
        long_description=packages.get_readme(),
        install_requires=packages.get_requirements("requirements/requirements.txt"),
        extras_require=extras_requirements,
        entry_points={
            "console_scripts": [
                "mypythontools_cicd = mypythontools_cicd.cicd:cicd_pipeline",
            ],
        },
        **packages.get_package_setup_args("mypythontools_cicd", development_status="alpha"),
        **packages.personal_setup_args_preset,
    )



"""Module with functions for 'git' subpackage."""

from __future__ import annotations
from typing import Sequence

from typing_extensions import Literal

from mypythontools.misc import print_progress
from mypythontools.types import validate_sequence
from mypythontools.system import (
    get_console_str_with_quotes,
    terminal_do_command,
    SHELL_AND,
)

# Lazy loaded
# from git import Repo

from ..project_paths import PROJECT_PATHS
from ..packages import get_version


def commit_all(commit_message: str, verbosity: Literal[0, 1, 2] = 1):
    """Stage all changes and create a commit.

    Args:
        commit_message (str): Commit message.
        verbosity (Literal[0, 1, 2], optional): If 0, prints nothing, if 1, then one line description of what
            happened is printed. If 3, all the results from terminal are printed. Defaults to 1.
    """
    print_progress("Creating commit of all changes", verbosity > 0)
    git_command = f"git add . {SHELL_AND} git commit -m {get_console_str_with_quotes(commit_message)}"
    terminal_do_command(git_command, cwd=PROJECT_PATHS.root.as_posix(), verbose=verbosity == 2)


def push(
    tag: str = "__version__",
    tag_message: str = "New version",
    verbosity: Literal[0, 1, 2] = 1,
) -> None:
    """Add tag and push.

    If tag is `__version__`, then tag is inferred from `__init__.py`.

    Args:
        tag (str, optional): Define tag used in push. If tag is '__version__', than is automatically generated
            from __init__ version. E.g from '1.0.2' to 'v1.0.2'.  Defaults to '__version__'.
        tag_message (str, optional): Message in annotated tag. Defaults to 'New version'.
        verbosity (Literal[0, 1, 2], optional): If 0, prints nothing, if 1, then one line description of what
            happened is printed. If 3, all the results from terminal are printed. Defaults to 1.
    """
    import git.repo
    import git.exc
    from git.exc import GitCommandError

    print_progress("Pushing to github", verbosity > 0)

    git_command = "git push"

    if tag == "__version__":
        tag = f"v{get_version()}"

    if tag:
        if not tag_message:
            tag_message = "New version"
        try:
            git.repo.Repo(PROJECT_PATHS.root.as_posix()).create_tag(tag, message=tag_message)
        except GitCommandError as err:
            raise RuntimeError("Tag creation failed. It can be because such a tag already exists.") from err

        git_command += " --follow-tags"

    try:
        terminal_do_command(git_command, cwd=PROJECT_PATHS.root.as_posix(), verbose=verbosity == 2)

    except RuntimeError as err:
        git.repo.Repo(PROJECT_PATHS.root.as_posix()).delete_tag(tag)  # type: ignore
        raise RuntimeError("Push to git failed. Version restored and created git tag deleted.") from err


def check_branch(allowed_branches: Sequence[str]):
    """Check if project is opened with one of particular branches anr raise error if not.

    Args:
        allowed_branches (Sequence[str]): Sequence of allowed branches.

    Raises:
        RuntimeError: If project is on branch that is not in allowed branches.
    """
    import git.repo
    from git.exc import InvalidGitRepositoryError

    validate_sequence(allowed_branches, "allowed_branches")

    try:
        branch = git.repo.Repo(PROJECT_PATHS.root.as_posix()).active_branch.name
    except InvalidGitRepositoryError:
        raise RuntimeError(
            "Loading of git project failed. Verify whether running pipeline from correct path. If "
            "checks branch with `allowed_branches', there has to be `.git` folder available."
        ) from None

    if branch not in allowed_branches:
        raise RuntimeError(
            "Pipeline started on branch that is not allowed."
            "If you want to use it anyway, add it to allowed_branches parameter and "
            "turn off changing version and creating tag."
        )
